import datetime


class UrlBuilder:
    def __init__(self, api_root):
        self.api_root = api_root

    def build_wk_url(self, endpoint, parameters=None, resource_id=None):
        if resource_id is not None:
            return "{0}{1}/{2}".format(self.api_root, endpoint, resource_id)
        else:
            parameter_string = self._build_query_parameters(parameters)
            return "{0}{1}{2}".format(self.api_root, endpoint, parameter_string)

    def _parse_parameter(self, parameter):
        key = parameter[0]
        value = parameter[1]
        if self._parameter_should_be_ignored(key, value):
            return None
        if isinstance(value, list):
            return "{}={}".format(key, ",".join(str(elem) for elem in value))
        elif isinstance(value, bool):
            return "{}={}".format(key, str(value).lower())
        elif isinstance(value, datetime.datetime):
            return "{}={}".format(key, value.isoformat())
        else:
            return "{}={}".format(key, str(value))

    def _parameter_should_be_ignored(self, key, value):
        return value is None or key in ["self", "resource_id", "fetch_all"]

    def _build_query_parameters(self, parameters):
        if parameters:
            query_parameters = list(
                map(self._parse_parameter, sorted(parameters.items()))
            )
            query_parameters = [qp for qp in query_parameters if qp is not None]
            if query_parameters:
                return "?{}".format("&".join(query_parameters))
        return ""



class InvalidWanikaniApiKeyException(Exception):
    """
    The client was initialized with an invalid V2 API key, causing
    Wanikani to return a ``401 unauthorized`` response.
    """

    pass


class UnknownResourceException(Exception):
    """
    The model factory was unable to determine what type of resource
    Wanikani is sending back, or is not familiar with it.
    """

    pass



USER_INFORMATION ={
  "object": "user",
  "url": "https://api.wanikani.com/v2/user",
  "data_updated_at": "2020-05-01T05:20:41.769053Z",
  "data": {
    "id": "7a18daeb-4067-4e77-b0ea-230c7c347ea8",
    "username": "Tadgh11",
    "level": 12,
    "profile_url": "https://www.wanikani.com/users/Tadgh11",
    "started_at": "2013-07-09T12:02:54.952786Z",
    "subscription": {
      "active": True,
      "type": "lifetime",
      "max_level_granted": 60,
      "period_ends_at": None
    },
    "current_vacation_started_at": None,
    "preferences": {
      "lessons_batch_size": 5,
      "lessons_autoplay_audio": True,
      "reviews_autoplay_audio": False,
      "lessons_presentation_order": "ascending_level_then_subject",
      "reviews_display_srs_indicator": True
    }
  }
}

SUBJECT = {
    "id": 2467,
    "object": "vocabulary",
    "url": "https://api.wanikani.com/v2/subjects/2467",
    "data_updated_at": "2018-05-21T21:52:43.041390Z",
    "data": {
        "created_at": "2012-02-28T08:04:47.000000Z",
        "level": 1,
        "slug": "一",
        "hidden_at": None,
        "document_url": "https://www.wanikani.com/vocabulary/%E4%B8%80",
        "characters": "一",
        "meanings": [{"meaning": "One", "primary": True, "accepted_answer": True}],
        "readings": [{"primary": True, "reading": "いち", "accepted_answer": True}],
        "parts_of_speech": ["numeral"],
        "component_subject_ids": [440],
        "auxiliary_meanings": [],
    },
}

SINGLE_SUBJECT = {
    "id": 1,
    "object": "radical",
    "url": "https://api.wanikani.com/v2/subjects/1",
    "data_updated_at": "2018-12-05T20:47:15.603277Z",
    "data": {
        "created_at": "2012-02-27T18:08:16.000000Z",
        "level": 1,
        "slug": "ground",
        "hidden_at": None,
        "document_url": "https://www.wanikani.com/radicals/ground",
        "characters": "一",
        "character_images": [
            {
                "url": "https://cdn.wanikani.com/images/legacy/1054-subject-1-normal-weight-black-original.png?1520987606",
                "metadata": {
                    "color": "#000000",
                    "dimensions": "1024x1024",
                    "style_name": "original",
                },
                "content_type": "image/png",
            }
        ],
        "meanings": [{"meaning": "Ground", "primary": True, "accepted_answer": True}],
        "auxiliary_meanings": [],
        "amalgamation_subject_ids": [2],
    },
}

EMPTY_SUBJECTS_PAGE = {
    "object": "collection",
    "url": "https://api.wanikani.com/v2/subjects?ids=1%2C2%2C3&slugs=abc%2C123&types=vocabulary",
    "pages": {"per_page": 1000, "next_url": None, "previous_url": None},
    "total_count": 0,
    "data_updated_at": None,
    "data": [],
}

SUBJECTS_PAGE = {
    "object": "collection",
    "url": "https://api.wanikani.com/v2/subjects",
    "pages": {"per_page": 1000, "next_url": None, "previous_url": None},
    "total_count": 3,
    "data_updated_at": "2018-07-05T22:22:07.129381Z",
    "data": [
        {
            "id": 3,
            "object": "radical",
            "url": "https://api.wanikani.com/v2/subjects/1",
            "data_updated_at": "2018-05-21T21:51:35.051365Z",
            "data": {
                "created_at": "2012-02-27T18:08:16.000000Z",
                "level": 1,
                "slug": "ground",
                "hidden_at": None,
                "document_url": "https://www.wanikani.com/radicals/ground",
                "characters": "一",
                "auxiliary_meanings": [],
                "character_images": [
                    {
                        "url": "https://cdn.wanikani.com/images/legacy/1054-subject-1-normal-weight-black-original.png?1520987606",
                        "metadata": {
                            "color": "#000000",
                            "dimensions": "1024x1024",
                            "style_name": "original",
                        },
                        "content_type": "image/png",
                    },
                    {
                        "url": "https://cdn.wanikani.com/images/legacy/1054-subject-1-normal-weight-black-1024px.png?1520987606",
                        "metadata": {
                            "color": "#000000",
                            "dimensions": "1024x1024",
                            "style_name": "1024px",
                        },
                        "content_type": "image/png",
                    },
                    {
                        "url": "https://cdn.wanikani.com/images/legacy/1054-subject-1-normal-weight-black-512px.png?1520987606",
                        "metadata": {
                            "color": "#000000",
                            "dimensions": "512x512",
                            "style_name": "512px",
                        },
                        "content_type": "image/png",
                    },
                    {
                        "url": "https://cdn.wanikani.com/images/legacy/1054-subject-1-normal-weight-black-256px.png?1520987606",
                        "metadata": {
                            "color": "#000000",
                            "dimensions": "256x256",
                            "style_name": "256px",
                        },
                        "content_type": "image/png",
                    },
                    {
                        "url": "https://cdn.wanikani.com/images/legacy/1054-subject-1-normal-weight-black-128px.png?1520987606",
                        "metadata": {
                            "color": "#000000",
                            "dimensions": "128x128",
                            "style_name": "128px",
                        },
                        "content_type": "image/png",
                    },
                    {
                        "url": "https://cdn.wanikani.com/images/legacy/1054-subject-1-normal-weight-black-64px.png?1520987606",
                        "metadata": {
                            "color": "#000000",
                            "dimensions": "64x64",
                            "style_name": "64px",
                        },
                        "content_type": "image/png",
                    },
                    {
                        "url": "https://cdn.wanikani.com/images/legacy/1054-subject-1-normal-weight-black-32px.png?1520987606",
                        "metadata": {
                            "color": "#000000",
                            "dimensions": "32x32",
                            "style_name": "32px",
                        },
                        "content_type": "image/png",
                    },
                    {
                        "url": "https://cdn.wanikani.com/images/legacy/576-subject-1-without-css-original.svg?1520987227",
                        "metadata": {"inline_styles": False},
                        "content_type": "image/svg+xml",
                    },
                    {
                        "url": "https://cdn.wanikani.com/images/legacy/98-subject-1-with-css-original.svg?1520987072",
                        "metadata": {"inline_styles": True},
                        "content_type": "image/svg+xml",
                    },
                ],
                "meanings": [
                    {"meaning": "Ground", "primary": True, "accepted_answer": True}
                ],
                "amalgamation_subject_ids": [
                    440,
                    449,
                    450,
                    451,
                    488,
                    531,
                    533,
                    568,
                    590,
                    609,
                    633,
                    635,
                    709,
                    710,
                    724,
                    783,
                    808,
                    913,
                    932,
                    965,
                    971,
                    1000,
                    1020,
                    1085,
                    1113,
                    1126,
                    1137,
                    1178,
                    1198,
                    1240,
                    1241,
                    1249,
                    1340,
                    1367,
                    1372,
                    1376,
                    1379,
                    1428,
                    1431,
                    1463,
                    1491,
                    1506,
                    1521,
                    1547,
                    1559,
                    1591,
                    1655,
                    1674,
                    1706,
                    1769,
                    1851,
                    1852,
                    1855,
                    1868,
                    1869,
                    1888,
                    1970,
                    2091,
                    2104,
                    2128,
                    2138,
                    2148,
                    2171,
                    2182,
                    2263,
                    2277,
                    2334,
                    2375,
                    2419,
                    2437,
                ],
            },
        },
        {
            "id": 1,
            "object": "kanji",
            "url": "https://api.wanikani.com/v2/subjects/534",
            "data_updated_at": "2018-05-21T21:51:48.658813Z",
            "data": {
                "created_at": "2012-03-02T02:11:55.000000Z",
                "level": 4,
                "slug": "央",
                "hidden_at": None,
                "document_url": "https://www.wanikani.com/kanji/%E5%A4%AE",
                "characters": "央",
                "meanings": [
                    {"meaning": "Center", "primary": True, "accepted_answer": True},
                    {"meaning": "Central", "primary": False, "accepted_answer": True},
                    {"meaning": "Centre", "primary": False, "accepted_answer": True},
                ],
                "readings": [
                    {
                        "type": "onyomi",
                        "primary": True,
                        "reading": "おう",
                        "accepted_answer": True,
                    }
                ],
                "auxiliary_meanings": [],
                "component_subject_ids": [29, 18],
                "amalgamation_subject_ids": [2726],
            },
        },
        {
            "id": 2,
            "object": "vocabulary",
            "url": "https://api.wanikani.com/v2/subjects/2467",
            "data_updated_at": "2018-05-21T21:52:43.041390Z",
            "data": {
                "created_at": "2012-02-28T08:04:47.000000Z",
                "level": 1,
                "slug": "一",
                "hidden_at": None,
                "document_url": "https://www.wanikani.com/vocabulary/%E4%B8%80",
                "characters": "一",
                "meanings": [
                    {"meaning": "One", "primary": True, "accepted_answer": True}
                ],
                "readings": [
                    {"primary": True, "reading": "いち", "accepted_answer": True}
                ],
                "parts_of_speech": ["numeral"],
                "auxiliary_meanings": [],
                "component_subject_ids": [440],
            },
        },
    ],
}

ASSIGNMENTS_PAGE = {
    "object": "collection",
    "url": "https://api.wanikani.com/v2/assignments",
    "pages": {
        "per_page": 500,
        "next_url": "https://api.wanikani.com/v2/assignments?page_after_id=62308815",
        "previous_url": None,
    },
    "total_count": 3,
    "data_updated_at": "2018-06-30T16:40:52.513654Z",
    "data": [
        {
            "id": 85899,
            "object": "assignment",
            "url": "https://api.wanikani.com/v2/assignments/85899",
            "data_updated_at": "2018-05-09T21:17:31.000000Z",
            "data": {
                "created_at": "2017-04-15T14:53:56.818837Z",
                "subject_id": 2,
                "subject_type": "vocabulary",
                "level": 5,
                "srs_stage": 9,
                "unlocked_at": "2017-04-15T14:53:56.818837Z",
                "started_at": "2017-04-15T14:53:56.818837Z",
                "passed_at": None,
                "burned_at": "2018-01-03T00:08:22.451866Z",
                "available_at": None,
                "resurrected_at": None,
                "passed": True,
                "hidden": False,
            },
        },
        {
            "id": 86555,
            "object": "assignment",
            "url": "https://api.wanikani.com/v2/assignments/86555",
            "data_updated_at": "2018-05-09T21:17:31.000000Z",
            "data": {
                "created_at": "2017-04-15T14:50:51.503084Z",
                "subject_id": 3,
                "subject_type": "vocabulary",
                "level": 5,
                "srs_stage": 9,
                "unlocked_at": "2017-04-15T14:50:51.503084Z",
                "started_at": "2017-04-15T14:50:51.503084Z",
                "passed_at": None,
                "burned_at": "2018-02-19T23:02:25.053105Z",
                "available_at": None,
                "resurrected_at": None,
                "passed": True,
                "hidden": False,
            },
        },
        {
            "id": 86606,
            "object": "assignment",
            "url": "https://api.wanikani.com/v2/assignments/86606",
            "data_updated_at": "2018-05-09T21:17:31.000000Z",
            "data": {
                "created_at": "2017-04-24T15:17:28.712677Z",
                "subject_id": 1,
                "subject_type": "vocabulary",
                "level": 6,
                "srs_stage": 9,
                "unlocked_at": "2017-04-24T15:17:28.712677Z",
                "started_at": "2017-04-24T15:17:28.712677Z",
                "passed_at": "2017-05-10T13:52:56.699204Z",
                "burned_at": "2018-02-19T22:46:09.144931Z",
                "available_at": None,
                "resurrected_at": None,
                "passed": True,
                "hidden": False,
            },
        },
    ],
}

REVIEW_STATISTICS_PAGE = {
    "object": "collection",
    "url": "https://api.wanikani.com/v2/review_statistics",
    "pages": {
        "per_page": 500,
        "next_url": "https://api.wanikani.com/v2/review_statistics?page_after_id=62308745",
        "previous_url": None,
    },
    "total_count": 5,
    "data_updated_at": "2018-05-24T22:02:41.393482Z",
    "data": [
        {
            "id": 85899,
            "object": "review_statistic",
            "url": "https://api.wanikani.com/v2/review_statistics/85899",
            "data_updated_at": "2018-01-03T00:08:22.469272Z",
            "data": {
                "created_at": "2017-04-15T14:53:56.818837Z",
                "subject_id": 1,
                "subject_type": "vocabulary",
                "meaning_correct": 13,
                "meaning_incorrect": 2,
                "meaning_max_streak": 7,
                "meaning_current_streak": 7,
                "reading_correct": 13,
                "reading_incorrect": 0,
                "reading_max_streak": 13,
                "reading_current_streak": 13,
                "percentage_correct": 93,
                "hidden": False,
            },
        },
        {
            "id": 86555,
            "object": "review_statistic",
            "url": "https://api.wanikani.com/v2/review_statistics/86555",
            "data_updated_at": "2018-02-19T23:02:25.114612Z",
            "data": {
                "created_at": "2017-04-15T14:50:51.503084Z",
                "subject_id": 2,
                "subject_type": "vocabulary",
                "meaning_correct": 11,
                "meaning_incorrect": 0,
                "meaning_max_streak": 11,
                "meaning_current_streak": 11,
                "reading_correct": 11,
                "reading_incorrect": 1,
                "reading_max_streak": 7,
                "reading_current_streak": 7,
                "percentage_correct": 96,
                "hidden": False,
            },
        },
        {
            "id": 86606,
            "object": "review_statistic",
            "url": "https://api.wanikani.com/v2/review_statistics/86606",
            "data_updated_at": "2018-02-19T22:46:09.166397Z",
            "data": {
                "created_at": "2017-04-24T15:17:28.712677Z",
                "subject_id": 3,
                "subject_type": "vocabulary",
                "meaning_correct": 8,
                "meaning_incorrect": 0,
                "meaning_max_streak": 8,
                "meaning_current_streak": 8,
                "reading_correct": 8,
                "reading_incorrect": 0,
                "reading_max_streak": 8,
                "reading_current_streak": 8,
                "percentage_correct": 100,
                "hidden": False,
            },
        },
        {
            "id": 86625,
            "object": "review_statistic",
            "url": "https://api.wanikani.com/v2/review_statistics/86625",
            "data_updated_at": "2018-02-19T23:54:40.912486Z",
            "data": {
                "created_at": "2017-04-24T15:17:29.061457Z",
                "subject_id": 1,
                "subject_type": "radical",
                "meaning_correct": 8,
                "meaning_incorrect": 0,
                "meaning_max_streak": 8,
                "meaning_current_streak": 8,
                "reading_correct": 1,
                "reading_incorrect": 0,
                "reading_max_streak": 1,
                "reading_current_streak": 1,
                "percentage_correct": 100,
                "hidden": False,
            },
        },
        {
            "id": 86891,
            "object": "review_statistic",
            "url": "https://api.wanikani.com/v2/review_statistics/86891",
            "data_updated_at": "2018-05-24T21:35:18.556752Z",
            "data": {
                "created_at": "2017-04-24T15:17:38.685804Z",
                "subject_id": 3,
                "subject_type": "vocabulary",
                "meaning_correct": 12,
                "meaning_incorrect": 1,
                "meaning_max_streak": 11,
                "meaning_current_streak": 11,
                "reading_correct": 12,
                "reading_incorrect": 1,
                "reading_max_streak": 9,
                "reading_current_streak": 3,
                "percentage_correct": 92,
                "hidden": False,
            },
        },
    ],
}

STUDY_MATERIALS_PAGE = {
    "object": "collection",
    "url": "https://api.wanikani.com/v2/study_materials",
    "pages": {"per_page": 500, "next_url": None, "previous_url": None},
    "total_count": 3,
    "data_updated_at": "2018-02-20T21:23:31.246408Z",
    "data": [
        {
            "id": 1539170,
            "object": "study_material",
            "url": "https://api.wanikani.com/v2/study_materials/1539170",
            "data_updated_at": "2017-06-01T19:01:36.573350Z",
            "data": {
                "created_at": "2017-02-01T15:55:42.058583Z",
                "subject_id": 7518,
                "subject_type": "vocabulary",
                "meaning_note": None,
                "reading_note": None,
                "meaning_synonyms": ["young girl"],
                "hidden": False,
            },
        },
        {
            "id": 1661853,
            "object": "study_material",
            "url": "https://api.wanikani.com/v2/study_materials/1661853",
            "data_updated_at": "2017-06-07T00:23:41.431508Z",
            "data": {
                "created_at": "2017-04-08T14:02:50.758641Z",
                "subject_id": 2798,
                "subject_type": "vocabulary",
                "meaning_note": None,
                "reading_note": None,
                "meaning_synonyms": ["balls"],
                "hidden": False,
            },
        },
        {
            "id": 1678472,
            "object": "study_material",
            "url": "https://api.wanikani.com/v2/study_materials/1678472",
            "data_updated_at": "2017-06-12T15:22:15.753065Z",
            "data": {
                "created_at": "2017-02-23T14:51:21.526934Z",
                "subject_id": 3416,
                "subject_type": "vocabulary",
                "meaning_note": None,
                "reading_note": None,
                "meaning_synonyms": ["wool"],
                "hidden": False,
            },
        },
    ],
}

SUMMARY = {
    "object": "report",
    "url": "https://api.wanikani.com/v2/summary",
    "data_updated_at": "2018-07-02T07:00:00.000000Z",
    "data": {
        "lessons": [{"available_at": "2018-07-02T07:00:00.000000Z", "subject_ids": []}],
        "next_reviews_at": "2018-07-02T09:00:00.000000Z",
        "reviews": [
            {"available_at": "2018-07-02T07:00:00.000000Z", "subject_ids": [1, 2, 3]},
            {"available_at": "2018-07-02T08:00:00.000000Z", "subject_ids": [4, 5, 6]},
            {"available_at": "2018-07-02T09:00:00.000000Z", "subject_ids": [647]},
            {"available_at": "2018-07-02T10:00:00.000000Z", "subject_ids": []},
            {"available_at": "2018-07-02T11:00:00.000000Z", "subject_ids": []},
            {"available_at": "2018-07-02T12:00:00.000000Z", "subject_ids": []},
            {
                "available_at": "2018-07-02T13:00:00.000000Z",
                "subject_ids": [8800, 2944, 2943],
            },
            {"available_at": "2018-07-02T14:00:00.000000Z", "subject_ids": []},
            {
                "available_at": "2018-07-02T15:00:00.000000Z",
                "subject_ids": [
                    658,
                    8738,
                    3447,
                    6237,
                    3449,
                    3451,
                    7676,
                    7528,
                    7621,
                    7679,
                    2822,
                    3420,
                    657,
                    5717,
                    3436,
                    7677,
                    7678,
                    3452,
                    7529,
                    3450,
                    3438,
                    7568,
                    7675,
                    3437,
                    3422,
                    3448,
                    4877,
                    7734,
                    7735,
                    666,
                    646,
                    648,
                ],
            },
            {"available_at": "2018-07-02T16:00:00.000000Z", "subject_ids": []},
            {"available_at": "2018-07-02T17:00:00.000000Z", "subject_ids": []},
            {"available_at": "2018-07-02T18:00:00.000000Z", "subject_ids": []},
            {"available_at": "2018-07-02T19:00:00.000000Z", "subject_ids": []},
            {"available_at": "2018-07-02T20:00:00.000000Z", "subject_ids": []},
            {"available_at": "2018-07-02T21:00:00.000000Z", "subject_ids": [2841]},
            {"available_at": "2018-07-02T22:00:00.000000Z", "subject_ids": []},
            {
                "available_at": "2018-07-02T23:00:00.000000Z",
                "subject_ids": [2945, 672, 2956, 2932, 2981, 2953, 674, 2936, 654],
            },
            {"available_at": "2018-07-03T00:00:00.000000Z", "subject_ids": []},
            {"available_at": "2018-07-03T01:00:00.000000Z", "subject_ids": []},
            {"available_at": "2018-07-03T02:00:00.000000Z", "subject_ids": []},
            {"available_at": "2018-07-03T03:00:00.000000Z", "subject_ids": []},
            {
                "available_at": "2018-07-03T04:00:00.000000Z",
                "subject_ids": [
                    671,
                    853,
                    3709,
                    2959,
                    4849,
                    2970,
                    2960,
                    2966,
                    2967,
                    2952,
                    2946,
                    8663,
                    2962,
                    2961,
                    2973,
                    2938,
                    2935,
                    2940,
                    7461,
                    2969,
                    2958,
                    2937,
                    7736,
                    2957,
                    8801,
                    2974,
                    677,
                    2939,
                    675,
                    663,
                    668,
                    650,
                    664,
                    670,
                    660,
                    676,
                ],
            },
            {"available_at": "2018-07-03T05:00:00.000000Z", "subject_ids": []},
            {"available_at": "2018-07-03T06:00:00.000000Z", "subject_ids": []},
            {"available_at": "2018-07-03T07:00:00.000000Z", "subject_ids": []},
        ],
    },
}
REVIEWS_PAGE = {
    "object": "collection",
    "url": "https://api.wanikani.com/v2/reviews",
    "pages": {
        "per_page": 1000,
        "next_url": "https://api.wanikani.com/v2/reviews?page_after_id=168707639",
        "previous_url": None,
    },
    "total_count": 3,
    "data_updated_at": "2018-07-06T19:30:19.657822Z",
    "data": [
        {
            "id": 6418820,
            "object": "review",
            "url": "https://api.wanikani.com/v2/reviews/6418820",
            "data_updated_at": "2017-08-13T14:32:50.580980Z",
            "data": {
                "created_at": "2017-08-13T14:32:50.580980Z",
                "assignment_id": 69392456,
                "subject_id": 2514,
                "starting_srs_stage": 8,
                "ending_srs_stage": 9,
                "incorrect_meaning_answers": 0,
                "incorrect_reading_answers": 0,
            },
        },
        {
            "id": 6418839,
            "object": "review",
            "url": "https://api.wanikani.com/v2/reviews/6418839",
            "data_updated_at": "2017-08-13T14:32:52.693772Z",
            "data": {
                "created_at": "2017-08-13T14:32:52.693772Z",
                "assignment_id": 30950170,
                "subject_id": 69,
                "starting_srs_stage": 8,
                "ending_srs_stage": 9,
                "incorrect_meaning_answers": 0,
                "incorrect_reading_answers": 0,
            },
        },
        {
            "id": 6418872,
            "object": "review",
            "url": "https://api.wanikani.com/v2/reviews/6418872",
            "data_updated_at": "2017-08-13T14:32:56.587244Z",
            "data": {
                "created_at": "2017-08-13T14:32:56.587244Z",
                "assignment_id": 30950168,
                "subject_id": 60,
                "starting_srs_stage": 8,
                "ending_srs_stage": 9,
                "incorrect_meaning_answers": 0,
                "incorrect_reading_answers": 0,
            },
        },
    ],
}

LEVEL_PROGRESSIONS_PAGE = {
    "object": "collection",
    "url": "https://api.wanikani.com/v2/level_progressions",
    "pages": {"per_page": 500, "next_url": None, "previous_url": None},
    "total_count": 2,
    "data_updated_at": "2018-07-05T18:03:21.967992Z",
    "data": [
        {
            "id": 15446,
            "object": "level_progression",
            "url": "https://api.wanikani.com/v2/level_progressions/15446",
            "data_updated_at": "2018-07-05T15:04:04.222661Z",
            "data": {
                "created_at": "2017-09-28T01:24:11.715238Z",
                "level": 7,
                "unlocked_at": "2017-06-12T15:24:48.181971Z",
                "started_at": "2017-09-28T01:24:11.707880Z",
                "passed_at": "2018-07-05T15:04:04.210181Z",
                "completed_at": None,
                "abandoned_at": None,
            },
        },
        {
            "id": 365549,
            "object": "level_progression",
            "url": "https://api.wanikani.com/v2/level_progressions/365549",
            "data_updated_at": "2018-07-05T18:03:21.967992Z",
            "data": {
                "created_at": "2018-07-05T15:04:04.365184Z",
                "level": 8,
                "unlocked_at": "2018-07-05T15:04:04.338492Z",
                "started_at": "2018-07-05T18:03:21.957917Z",
                "passed_at": None,
                "completed_at": None,
                "abandoned_at": None,
            },
        },
    ],
}

RESETS_PAGE = {
    "object": "collection",
    "url": "https://api.wanikani.com/v2/resets",
    "pages": {"per_page": 500, "next_url": None, "previous_url": None},
    "total_count": 1,
    "data_updated_at": "2018-03-21T22:07:39.261116Z",
    "data": [
        {
            "id": 6529,
            "object": "reset",
            "url": "https://api.wanikani.com/v2/resets/6529",
            "data_updated_at": "2018-03-21T22:07:39.261116Z",
            "data": {
                "created_at": "2018-03-21T22:04:13.313903Z",
                "original_level": 13,
                "target_level": 1,
                "confirmed_at": "2018-03-21T22:05:44.454026Z",
            },
        }
    ],
}



import requests

from wanikani_api import constants, models
from wanikani_api.exceptions import InvalidWanikaniApiKeyException
from wanikani_api.models import Iterator
from wanikani_api.subjectcache import SubjectCache
from wanikani_api.url_builder import UrlBuilder
from copy import deepcopy


"""
This is the primary point of entry for accessing the Wanikani API.
"""


class Client:
    """
    This is the only object you can instantiate. It provides access to each
    relevant API endpoint on Wanikani.
    """

    def __init__(self, v2_api_key, subject_cache_enabled=False):
        self.v2_api_key = v2_api_key
        self.headers = {"Authorization": "Bearer {}".format(v2_api_key)}
        self.url_builder = UrlBuilder(constants.ROOT_WK_API_URL)
        self.subject_cache = None
        self.etag_cache = {}
        self.authorized_request_maker = self.build_authorized_requester(self.headers)
        if subject_cache_enabled:
            self.use_local_subject_cache()

    def _fetch_result_from_cache(self, request_key):
        return self.etag_cache[request_key]["result"]

    def _fetch_etag_from_cache(self, request_key):
        return self.etag_cache[request_key]["etag"]

    def _store_in_cache(self, request_key, response):
        etag = response.headers["Etag"]
        self.etag_cache[request_key]["etag"] = etag
        self.etag_cache[request_key]["result"] = self._serialize_wanikani_response(
            response
        )

    def build_authorized_requester(self, headers):
        def _make_wanikani_api_request(url):
            request_key = (url, headers["Authorization"])
            request_headers = deepcopy(headers)
            try:
                etag = self.etag_cache[request_key]["etag"]
                request_headers["If-None-Match"] = etag
            except KeyError:
                self.etag_cache[request_key] = {}
            finally:
                response = requests.get(url, headers=request_headers)
                if response.status_code == 304:
                    return self._fetch_result_from_cache(request_key)
                elif response.status_code == 200:
                    self._store_in_cache(request_key, response)
                    return self.etag_cache[request_key]["result"]
                else:
                    raise Exception(f"Failed to contact Wanikani: {response.content}")

        return _make_wanikani_api_request

    def use_local_subject_cache(self):
        self.subject_cache = SubjectCache(self.subjects(fetch_all=True))

    def user_information(self):
        """
        Gets all relevant information about the user.

        :raises: :class:`.exceptions.InvalidWanikaniApiKeyException`
        :rtype: :class:`.models.UserInformation`
        """
        response = requests.get(
            self.url_builder.build_wk_url(constants.USER_ENDPOINT), headers=self.headers
        )
        return self._serialize_wanikani_response(response)

    def subject(self, subject_id):
        """
        Get a single subject by its known id

        :param subject_id: the id of the subject
        :return: a single :class:`.models.Subject`. This might be either:
            * :class:`.models.Radical`
            * :class:`.models.Kanji`
            * :class:`.models.Vocabulary`
        """
        if self.subject_cache:
            return self.subject_cache.get(subject_id)
        response = requests.get(
            self.url_builder.build_wk_url(
                constants.SUBJECT_ENDPOINT, resource_id=subject_id
            ),
            headers=self.headers,
        )
        return self._serialize_wanikani_response(response)

    def subjects(
        self,
        ids=None,
        types=None,
        slugs=None,
        levels=None,
        hidden=None,
        updated_after=None,
        fetch_all=False,
    ):
        """Retrieves Subjects

        Wanikani refers to Radicals, Kanji, and Vocabulary as Subjects. This function allows you to fetch all of
        the subjects, regardless of the current level of the account that the API key is associated to. All parameters
        to this function are optional, and are for filtering the results.
        are ignored, and the subject with that ID in question is fetched.

        :param int[] ids: Filters based on a list of IDs. Does not cause other parameters to be ignored.
        :param str[] types: The specific :class:`.models.Subject` types you wish to retrieve. Possible values are: ``["kanji", "vocabulary", "radicals"]``
        :param str[] slugs: The wanikani slug
        :param int[] levels: Include only :class:`.models.Subject` from the specified levels.
        :param bool hidden: Return :class:`.models.Subject` which are or are not hidden from the user-facing application
        :param bool fetch_all: if set to True, instead of fetching only first page of results, will fetch them all.
        :param  updated_after: Return results which have been updated after the timestamp
        :type updated_after: :class:`datetime.datetime`
        :return: An iterator over multiple :class:`models.Page` , in which the ``data`` field contains a list anything that is a :class:`.models.Subject`, e.g.:

            * :class:`.models.Radical`
            * :class:`.models.Kanji`
            * :class:`.models.Vocabulary`
        """

        url = self.url_builder.build_wk_url(
            constants.SUBJECT_ENDPOINT, parameters=locals()
        )
        return self._wrap_collection_in_iterator(
            self.authorized_request_maker(url), fetch_all
        )

    def assignment(self, assignment_id):
        """
        Get a single :class:`.models.Assignment` by its known id

        :param assignment_id: the id of the assignment
        :return: a single :class:`.models.Assignment`
        """
        response = requests.get(
            self.url_builder.build_wk_url(
                constants.ASSIGNMENT_ENDPOINT, resource_id=assignment_id
            ),
            headers=self.headers,
        )
        return self._serialize_wanikani_response(response)

    def assignments(
        self,
        ids=None,
        created_at=None,
        subject_ids=None,
        subject_types=None,
        levels=None,
        available_before=None,
        available_after=None,
        srs_stages=None,
        unlocked=None,
        started=None,
        passed=None,
        burned=None,
        resurrected=None,
        hidden=None,
        updated_after=None,
        fetch_all=False,
    ):
        """
        Assignments are the association between a user, and a subject. This means that every time something is added to
        your lessons, a new :class:`.models.Assignment` is created.

        :param bool fetch_all: if set to True, instead of fetching only first page of results, will fetch them all.
        :param int[] ids: Return only results with the given IDs
        :param created_at: Timestamp when resource was created
        :param int[] subject_ids: Return only :class:`.models.Assignment`s which are tied to the given subject_ids
        :param str[] subject_types: The specific :class:`.models.Subject` types you wish to retrieve. Possible values are: ``["kanji", "vocabulary", "radicals"]``
        :param int[] levels: Include only :class:`.models.Assignment` where the subjects are from the specified levels.
        :param datetime available_before: Return assignment reviews available before timestamp
        :param datetime available_after: Return assignment reviews available after timestamp
        :param int srs_stages: Return assignments of specified srs stages. Note, 0 is lessons, 9 is the burned state
        :param bool unlocked: Return assignments which have unlocked (made available to lessons)
        :param bool started: Return assignments which move from lessons to reviews
        :param bool passed: Return assignments which have reach Guru (aka srs_stage 5) at some point (true) or which have never been Guru’d (false)
        :param bool burned: Return assignments which have been burned at some point (true) or never have been burned (false)
        :param bool resurrected: Return assignments which either have been resurrect (true) or not (false)
        :param bool hidden: Return assignments which are or are not hidden from the user-facing application
        :param datetime updated_after: Return results which have been updated after the timestamp
        :return: An iterator over a set of :class:`.models.Page` where the data contained is all :class:`.models.Assignment`
        """
        response = requests.get(
            self.url_builder.build_wk_url(
                constants.ASSIGNMENT_ENDPOINT, parameters=locals()
            ),
            headers=self.headers,
        )
        return self._wrap_collection_in_iterator(
            self._serialize_wanikani_response(response), fetch_all
        )

    def review_statistic(self, review_statistic_id):
        """
        Get a single :class:`.models.ReviewStatistic` by its known id

        :param review_statistic_id: the id of the review_statistic
        :return: a single :class:`.models.ReviewStatistic`
        """
        response = requests.get(
            self.url_builder.build_wk_url(
                constants.REVIEW_STATS_ENDPOINT, resource_id=review_statistic_id
            ),
            headers=self.headers,
        )
        return self._serialize_wanikani_response(response)

    def review_statistics(
        self,
        ids=None,
        subject_ids=None,
        subject_types=None,
        updated_after=None,
        percentages_greater_than=None,
        percentages_less_than=None,
        hidden=None,
        fetch_all=False,
    ):
        """
        Retrieve all Review Statistics from Wanikani. A Review Statistic is related to a single subject which the user has studied.

        :param bool fetch_all: if set to True, instead of fetching only first page of results, will fetch them all.
        :param int[] ids: Return only results with the given IDs
        :param int[] subject_ids: Return only :class:`.models.Assignment`s which are tied to the given subject_ids
        :param str[] subject_types: The specific :class:`.models.Subject` types you wish to retrieve. Possible values are: ``["kanji", "vocabulary", "radicals"]``
        :param datetime updated_after: Return results which have been updated after the timestamp
        :param int percentages_greater_than: Return results where the percentage_correct is greater than the value. [0-100]
        :param int percentages_less_than: Return results where the percentage_correct is less than the value. [0-100]
        :param bool hidden: Return only results where the related subject has been hidden.
        :return: An iterator which contains all Review Statistics
        """
        response = requests.get(
            self.url_builder.build_wk_url(
                constants.REVIEW_STATS_ENDPOINT, parameters=locals()
            ),
            headers=self.headers,
        )
        return self._wrap_collection_in_iterator(
            self._serialize_wanikani_response(response), fetch_all
        )

    def study_material(self, study_material_id):
        """
        Get a single :class:`.models.StudyMaterial` by its known id

        :param study_material_id: the id of the study material
        :return: a single :class:`.models.StudyMaterial`
        """
        response = requests.get(
            self.url_builder.build_wk_url(
                constants.STUDY_MATERIALS_ENDPOINT, resource_id=study_material_id
            ),
            headers=self.headers,
        )
        return self._serialize_wanikani_response(response)

    def study_materials(
        self,
        ids=None,
        subject_ids=None,
        subject_types=None,
        hidden=None,
        updated_after=None,
        fetch_all=False,
    ):
        """
        Retrieve all Study Materials. These are primarily meaning notes, reading notes, and meaning synonyms.

        :param bool fetch_all: if set to True, instead of fetching only first page of results, will fetch them all.
        :param int[] ids: Return only results with the given IDs
        :param int[] subject_ids: Return only :class:`.models.Assignment`s which are tied to the given subject_ids
        :param str[] subject_types: The specific :class:`.models.Subject` types you wish to retrieve. Possible values are: ``["kanji", "vocabulary", "radicals"]``
        :param bool hidden: Return only results where the related subject has been hidden.
        :param datetime updated_after: Return results which have been updated after the timestamp
        :return: An iterator over all Study Materials
        """
        response = requests.get(
            self.url_builder.build_wk_url(
                constants.STUDY_MATERIALS_ENDPOINT, parameters=locals()
            ),
            headers=self.headers,
        )
        return self._wrap_collection_in_iterator(
            self._serialize_wanikani_response(response), fetch_all
        )

    def summary(self):
        """

        :return:
        """
        response = requests.get(
            self.url_builder.build_wk_url(
                constants.SUMMARY_ENDPOINT, parameters=locals()
            ),
            headers=self.headers,
        )
        return self._serialize_wanikani_response(response)

    def review(self, review_id):
        """
        Get a single :class:`.models.Review` by its known id

        :param review_id: the id of the review
        :return: a single :class:`.models.Review`
        """
        response = requests.get(
            self.url_builder.build_wk_url(
                constants.REVIEWS_ENDPOINT, resource_id=review_id
            ),
            headers=self.headers,
        )
        return self._serialize_wanikani_response(response)

    def reviews(self, ids=None, subject_ids=None, updated_after=None, fetch_all=False):
        """
        Retrieve all reviews for a given user. A :class:`.models.Review` is a single instance of this user getting a
        single review correctly submitted.

        :param bool fetch_all: if set to True, instead of fetching only first page of results, will fetch them all.
        :param int[] ids: Return only results with the given IDs
        :param int[] subject_ids: Return only :class:`.models.Assignment`s which are tied to the given subject_ids
        :param datetime updated_after: Return results which have been updated after the timestamp
        :return: An iterator over all :class:`.models.Review` for a given user.
        """
        response = requests.get(
            self.url_builder.build_wk_url(
                constants.REVIEWS_ENDPOINT, parameters=locals()
            ),
            headers=self.headers,
        )
        return self._wrap_collection_in_iterator(
            self._serialize_wanikani_response(response), fetch_all
        )

    def level_progression(self, level_progression_id):
        """
        Get a single :class:`.models.LevelProgression` by its known id

        :param level_progression_id: the id of the level_progression
        :return: a single :class:`.models.LevelProgression`
        """
        url = self.url_builder.build_wk_url(
            constants.LEVEL_PROGRESSIONS_ENDPOINT, resource_id=level_progression_id
        )
        return self.authorized_request_maker(url)

    def level_progressions(self, ids=None, updated_after=None, fetch_all=False):
        """
        Retrieve all :class:`.models.LevelProgression` for a given user.

        :param bool fetch_all: if set to True, instead of fetching only first page of results, will fetch them all.
        :param int[] ids: Return only results with the given IDs
        :param datetime updated_after: Return results which have been updated after the timestamp
        :return: An iterator over all :class:`.models.LevelProgression` for a given user.
        """
        url =  self.url_builder.build_wk_url(
                constants.LEVEL_PROGRESSIONS_ENDPOINT, parameters=locals()
            )

        return self._wrap_collection_in_iterator(
            self.authorized_request_maker(url), fetch_all
        )

    def reset(self, reset_id):
        """
        Get a single :class:`.models.Reset` by its known id

        :param reset_id: the id of the reset
        :return: a single :class:`.models.Reset`
        """
        response = requests.get(
            self.url_builder.build_wk_url(
                constants.REVIEWS_ENDPOINT, resource_id=reset_id
            ),
            headers=self.headers,
        )
        return self._serialize_wanikani_response(response)

    def resets(self, ids=None, updated_after=None, fetch_all=False):
        """
        Retrieve information for all resets the user has performed on Wanikani.

        :param bool fetch_all: if set to True, instead of fetching only first page of results, will fetch them all.
        :param int[] ids: Return only results with the given IDs
        :param datetime updated_after: Return results which have been updated after the timestamp
        :return: An iterator over all :class:`.models.Reset` for a given user.
        """
        response = requests.get(
            self.url_builder.build_wk_url(
                constants.RESETS_ENDPOINT, parameters=locals()
            ),
            headers=self.headers,
        )
        return self._wrap_collection_in_iterator(
            self._serialize_wanikani_response(response), fetch_all
        )

    def _serialize_wanikani_response(self, response):
        if response.status_code == 200:
            json = response.json()
            return models.factory(json, client=self)
        elif response.status_code == 401:
            raise InvalidWanikaniApiKeyException(
                "[{}] is not a valid API key!".format(self.v2_api_key)
            )

    def _wrap_collection_in_iterator(self, resource, fetch_all):
        return Iterator(
            current_page=resource,
            api_request=self.authorized_request_maker,
            fetch_all=fetch_all,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `wanikani_api` package."""
import requests

from wanikani_api.client import Client
from wanikani_api.models import Subject, UserInformation
from tests.utils.utils import (
    mock_user_info,
    mock_subjects,
    mock_assignments,
    mock_review_statistics,
    mock_study_materials,
    mock_summary,
    mock_reviews,
    mock_level_progressions,
    mock_resets,
    mock_single_subject,
)


def test_subjectable_mixin_works(requests_mock):
    mock_subjects(requests_mock)
    mock_assignments(requests_mock)
    client = Client("arbitrary_api_key", subject_cache_enabled=True)

    assignments = client.assignments()
    assignment = assignments[0]
    subj = assignment.subject
    assert subj.id == assignment.subject_id


def test_expected_subjectable_resources_work(requests_mock):
    mock_assignments(requests_mock)
    mock_summary(requests_mock)
    mock_subjects(requests_mock)
    mock_reviews(requests_mock)
    mock_single_subject(requests_mock)
    mock_study_materials(requests_mock)
    mock_summary(requests_mock)

    client = Client("arbitrary_api_key")

    assignments = client.assignments()
    assert assignments[0].subject is not None

    study_materials = client.study_materials()
    assert study_materials[0].subject is not None

    study_materials = client.study_materials()
    assert study_materials[0].subject is not None

    summary = client.summary()
    assert summary.reviews[0].subjects is not None

    reviews = client.reviews()
    assert reviews[0].subject is not None




ROOT_WK_API_URL = "https://api.wanikani.com/v2/"
RESOURCES_WITHOUT_IDS = ["user", "collection", "report"]
SUBJECT_ENDPOINT = "subjects"
SINGLE_SUBJECT_ENPOINT = r"subjects/\d+"
ASSIGNMENT_ENDPOINT = "assignments"
REVIEW_STATS_ENDPOINT = "review_statistics"
STUDY_MATERIALS_ENDPOINT = "study_materials"
REVIEWS_ENDPOINT = "reviews"
LEVEL_PROGRESSIONS_ENDPOINT = "level_progressions"
RESETS_ENDPOINT = "resets"
SUMMARY_ENDPOINT = "summary"
USER_ENDPOINT = "user"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `wanikani_api` package."""
import datetime

import requests
from tests.utils.utils import mock_subjects, mock_empty_subjects

from wanikani_api.client import Client
from wanikani_api.models import Iterator


class Empty200:
    def __init__(self):
        self.status_code = 200


class MockedRequest:
    def __init__(self, *args, **kwargs):
        self.status_code = 200


def test_subject_parameters_are_properly_converted(requests_mock):
    mock_subjects(requests_mock)

    v2_api_key = "arbitrary_api_key"
    client = Client(v2_api_key)

    client.subjects(ids=[1, 2, 3], hidden=False, slugs=["abc", "123"])

    assert requests_mock.call_count == 1
    assert (
        requests_mock.request_history[0].url
        == "https://api.wanikani.com/v2/subjects?hidden=false&ids=1,2,3&slugs=abc,123"
    )


def test_client_correctly_renders_empty_collections(requests_mock):
    mock_empty_subjects(requests_mock)
    v2_api_key = "arbitrary_api_key"
    client = Client(v2_api_key)
    response = client.subjects(ids=[1, 2, 3], hidden=False, slugs=["abc", "123"])
    assert len(response.current_page.data) == 0


def test_parameters_convert_datetime_to_string_correctly(requests_mock):
    mock_subjects(requests_mock)
    v2_api_key = "arbitrary_api_key"
    client = Client(v2_api_key)
    now = datetime.datetime.now()

    client.subjects(updated_after=now)

    assert requests_mock.call_count == 1
    assert (
        requests_mock.request_history[0].url
        == "https://api.wanikani.com/v2/subjects?updated_after=" + now.isoformat()
    )


def test_requests_mock(requests_mock):
    mock_subjects(requests_mock)

    client = Client("whatever")
    subjects = client.subjects()
    assert isinstance(subjects, Iterator)



class SubjectCache:
    class __SubjectCache:
        def __init__(self, subjects):

            self._populate_cache(subjects)

        def get(self, subject_id):
            try:
                return self.cached_subjects[subject_id]
            except KeyError:
                raise KeyError(
                    "We couldn't find a subject with ID {} in the cache!".format(
                        subject_id
                    )
                )

        def _populate_cache(self, subjects):
            self.cached_subjects = {subject.id: subject for subject in subjects}

    instance = None

    def __init__(self, subjects):
        if not SubjectCache.instance:
            SubjectCache.instance = SubjectCache.__SubjectCache(subjects)

    def __getattr__(self, name):
        return getattr(self.instance, name)



import functools
import operator
import pprint

import dateutil.parser

from wanikani_api import constants
from wanikani_api.exceptions import UnknownResourceException


class Resource:
    def __init__(self, json_data, *args, **kwargs):
        self.resource = json_data["object"]
        self._raw = json_data
        self.url = json_data["url"]
        self.data_updated_at = parse8601(json_data["data_updated_at"])
        # Some Resources do not have IDs.
        self.id = (
            None
            if self.resource in constants.RESOURCES_WITHOUT_IDS
            else json_data["id"]
        )
        self._resource = json_data["data"]

    def raw_json(self):
        return pprint.pformat(self._raw)


class Iterator:
    def __init__(self, current_page, api_request, max_results=None, fetch_all=False):
        self.current_page = current_page
        self.api_request = api_request
        self.max_results = max_results
        self.yielded_count = 0
        self.pages = [current_page]
        if current_page:
            self.per_page = current_page.per_page
        if fetch_all:
            self.fetch_all_pages()

    def _keep_iterating(self):
        return (
            self.current_page is not None
            and self.max_results
            and self.yielded_count >= self.max_results
        )

    def fetch_next_page(self):
        if self.current_page.next_page_url is not None:
            self.pages.append(self.api_request(self.current_page.next_page_url))
            self.current_page = self.pages[-1]

    def fetch_all_pages(self):
        while self.current_page.next_page_url is not None:
            self.fetch_next_page()

    def __iter__(self):
        return iter([item for page in self.pages for item in page])

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.pages[item // self.per_page][item % self.per_page]
        if isinstance(item, slice):
            return [self[i] for i in range(*item.indices(len(self)))]

    def __len__(self):
        return functools.reduce(operator.add, [len(page) for page in self.pages])


class Page(Resource):
    resource = "collection"

    def __init__(self, json_data, *args, **kwargs):
        super().__init__(json_data, *args, **kwargs)
        self.client = kwargs.get("client")
        self.next_page_url = json_data["pages"]["next_url"]
        self.previous_page_url = json_data["pages"]["previous_url"]
        self.total_count = json_data["total_count"]
        self.per_page = json_data["pages"]["per_page"]
        self.data = [factory(datum, client=self.client) for datum in json_data["data"]]

    def __iter__(self):
        return iter(self.data)

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.data[item]
        elif isinstance(item, slice):
            return [self.data[i] for i in range(*item.indices(len(self)))]

    def __len__(self):
        return len(self.data)


class Subjectable:
    """
    A Mixin allowing a model to quickly fetch related subjects.

    Any resource which inherits Subjectable must have either `subject_id` or `subject_ids` as an attribute.
    """

    def __init__(self, *args, **kwargs):
        if "client" not in kwargs:
            raise ValueError("Subjectable models expect an instance of Client!")
        self.client = kwargs.get("client")
        self._subjects = None
        self._subject = None

    @property
    def subject(self):
        if self._subject:
            return self._subject
        elif hasattr(self, "subject_id"):
            self._subject = self.client.subject(self.subject_id)
            return self._subject
        else:
            raise AttributeError("no attribute named subject!")

    @property
    def subjects(self):
        if self._subjects:
            return self._subjects
        elif hasattr(self, "subject_ids"):
            self._subjects = self.client.subjects(ids=self.subject_ids)
            return self._subjects
        else:
            raise AttributeError("no attribute named subjects!")


class Preferences():
    def __init__(self, preferences_json):
        self.default_voice_actor_id = preferences_json.get("default_voice_actor_id")
        self.lessons_batch_size= preferences_json["lessons_batch_size"]
        self.lessons_autoplay_audio = preferences_json["lessons_autoplay_audio"]
        self.reviews_autoplay_audio = preferences_json["reviews_autoplay_audio"]
        self.lessons_presentation_order = preferences_json["lessons_presentation_order"]
        self.reviews_display_srs_indicator = preferences_json["reviews_display_srs_indicator"]

class UserInformation(Resource):
    """
    This is a simple container for information returned from the ``/user/`` endpoint. This is all information related to
    the user.
    """

    resource = "user"

    def __init__(self, json_data, *args, **kwargs):
        super().__init__(json_data, *args, **kwargs)
        self.username = self._resource["username"]  #: username
        self.level = self._resource["level"]  #: current wanikani level
        self.subscription = Subscription(self._resource['subscription']) #: maximum level granted by subscription.
        self.profile_url = self._resource["profile_url"]  #: Link to user's profile.
        self.started_at = parse8601(
            self._resource["started_at"]
        )  #: datetime at which the user signed up.
        self.current_vacation_started_at = parse8601(
            self._resource["current_vacation_started_at"]
        )  #: datetime at which vacation was enabled on wanikani.
        self.preferences = Preferences(self._resource["preferences"])

    def __str__(self):
        return "UserInformation{{ username:{}, level:{}, profile_url:{} started_at:{}, current_vacation_started_at:{} }}".format(
            self.username,
            self.level,
            self.profile_url,
            self.started_at,
            self.current_vacation_started_at,
        )


class Subject(Resource):
    """
    This is the base Subject for Wanikani. This contains information common to Kanji, Vocabulary, and Radicals.
    """

    def __init__(self, json_data, *args, **kwargs):
        super().__init__(json_data, *args, **kwargs)
        resource_data = json_data["data"]
        self.level = resource_data["level"]  #: The level of the subject.
        self.created_at = parse8601(
            resource_data["created_at"]
        )  #: The date at which the Subject was created originally on Wanikani.
        self.characters = resource_data[
            "characters"
        ]  #: The actual japanese kanji/radical symbol such as 女
        self.meanings = [
            Meaning(meaning_json) for meaning_json in resource_data["meanings"]
        ]  #: A list of :class:`.models.Meaning` for this subject.
        self.document_url = resource_data[
            "document_url"
        ]  #: The direct URL where the subject can be found on Wanikani
        self.hidden_at = resource_data[
            "hidden_at"
        ]  #: When Wanikani removes a subject, they seem to instead set it to hidden, for backwards compatibilty with clients.
        self.auxiliary_meanings = [
            AuxiliaryMeaning(auxiliary_meaning_json) for auxiliary_meaning_json in resource_data["auxiliary_meanings"]
        ]

    def __str__(self) -> str:
        return f"{['['+meaning.meaning+']' if meaning.primary else meaning.meaning for meaning in self.meanings]}:{[character for character in self.characters] if self.characters else 'UNAVAILABLE'}"


class Radical(Subject):
    """
    A model for the Radical object.
    """

    resource = "radical"

    def __init__(self, json_data, *args, **kwargs):
        super().__init__(json_data, *args, **kwargs)
        self.character_images = (
            self._resource["character_images"]
            if "character_images" in self._resource.keys()
            else None
        )  #: A list of dictionaries, each containing a bunch of information related to a single character image.
        self.amalgamation_subject_ids = self._resource[
            "amalgamation_subject_ids"
        ]  #: IDs for various other :class:`.models.Subject` for which this radical is a component.

    def __str__(self) -> str:
        return f"Radical: {[meaning.meaning for meaning in self.meanings]}:{[character for character in self.characters] if self.characters else 'UNAVAILABLE'}"


class Vocabulary(Subject):
    """
    A model for the Vocabulary Resource
    """

    resource = "vocabulary"

    def __init__(self, json_data, *args, **kwargs):
        super().__init__(json_data, *args, **kwargs)
        self.parts_of_speech = self._resource[
            "parts_of_speech"
        ]  #: A list of strings, each of which is a part of speech.
        self.component_subject_ids = self._resource[
            "component_subject_ids"
        ]  #: List of IDs for :class"`.models.Kanji` which make up this vocabulary.
        self.readings = [
            Reading(reading_json) for reading_json in self._resource["readings"]
        ]  #: A list of :class:`.models.Reading` related to this Vocabulary.

    def __str__(self):
        return f"Vocabulary: {super(Vocabulary, self).__str__()}"


class Kanji(Subject):
    """
    A model for the Kanji Resource
    """

    resource = "kanji"

    def __init__(self, json_data, *args, **kwargs):
        super().__init__(json_data, *args, **kwargs)
        self.amalgamation_subject_ids = self._resource[
            "amalgamation_subject_ids"
        ]  #: A list of IDs for the related :class:`.models.Vocabulary` which this Kanji is a component in.
        self.component_subject_ids = self._resource[
            "component_subject_ids"
        ]  #: A list of IDs for the related :class:`.models.Radical` which combine to make this kanji
        self.readings = [
            Reading(reading_json) for reading_json in self._resource["readings"]
        ]  #: A list of :class:`.models.Reading` related to this Vocabulary.

    def __str__(self):
        return f"Kanji: {super(Kanji, self).__str__()}"

class AuxiliaryMeaning:
    """
    Simple data class for handling auxiliary meanings
    """
    def __init__(self, auxiliary_meaning_json):
        self.meaning = auxiliary_meaning_json["meaning"]
        self.type = auxiliary_meaning_json["type"]

    def __str__(self) -> str:
        return f"{self.meaning}({type})"

class Subscription:
    def __init__(self, subscription_json):
        self.active = subscription_json["active"]
        self.type = subscription_json["type"]
        self.max_level_granted = subscription_json["max_level_granted"]
        self.period_ends_at = parse8601(subscription_json["period_ends_at"])

    def __str__(self) -> str:
        return f"{'active' if self.active else 'inactive'}:{self.type}:{self.max_level_granted}:{self.period_ends_at}"

class Meaning:
    """
    Simple class holding information about a given meaning of a vocabulary/Kanji
    """

    def __init__(self, meaning_json):
        self.meaning = meaning_json["meaning"]  #: The english meaning of a Subject.
        self.primary = meaning_json[
            "primary"
        ]  #: Wether or not the meaning is considered to be the main one.
        self.accepted_answer = meaning_json[
            "accepted_answer"
        ]  #: Whether or not this answer is accepted during reviews in Wanikani.

    def __str__(self) -> str:
        return self.meaning


class Reading:
    """
    Simple class holding information about a given reading of a vocabulary/kanji
    """

    def __init__(self, meaning_json):
        #: the actual かな for the reading.
        self.reading = meaning_json["reading"]
        self.primary = meaning_json["primary"]  #: Whether this is the primary reading.
        self.accepted_answer = meaning_json[
            "accepted_answer"
        ]  #: Whether this answer is accepted as correct by Wanikani during review.


class Assignment(Resource, Subjectable):
    """
    Simple class holding information about Assignmetns.
    """

    resource = "assignment"

    def __init__(self, json_data, *args, **kwargs):
        super().__init__(json_data, *args, **kwargs)
        Subjectable.__init__(self, *args, **kwargs)
        self.created_at = parse8601(self._resource["created_at"])
        self.subject_id = self._resource["subject_id"]
        self.subject_type = self._resource["subject_type"]
        self.srs_stage = self._resource["srs_stage"]
        self.unlocked_at = parse8601(self._resource["unlocked_at"])
        self.started_at = parse8601(self._resource["started_at"])
        self.passed_at = parse8601(self._resource["passed_at"])
        self.burned_at = parse8601(self._resource["burned_at"])
        self.available_at = parse8601(self._resource["available_at"])
        self.resurrected_at_at = parse8601(self._resource["resurrected_at"])
        self.hidden = self._resource["hidden"]


class Reset(Resource):
    """
    Simple model holding resource information
    """

    resource = "reset"

    def __init__(self, json_data, *args, **kwargs):
        super().__init__(json_data, *args, **kwargs)
        self.created_at = parse8601(self._resource["created_at"])
        self.original_level = self._resource["original_level"]
        self.target_level = self._resource["target_level"]
        self.confirmed_at = parse8601(self._resource["confirmed_at"])


class ReviewStatistic(Resource):
    """
    Simple model holding ReviewStatistic Information
    """

    resource = "review_statistic"

    def __init__(self, json_data, *args, **kwargs):
        super().__init__(json_data, *args, **kwargs)
        self.created_at = parse8601(self._resource["created_at"])
        self.subject_id = self._resource["subject_id"]
        self.subject_type = self._resource["subject_type"]
        self.meaning_correct = self._resource["meaning_correct"]
        self.meaning_incorrect = self._resource["meaning_incorrect"]
        self.meaning_max_streak = self._resource["meaning_max_streak"]
        self.meaning_current_streak = self._resource["meaning_current_streak"]
        self.reading_correct = self._resource["reading_correct"]
        self.reading_incorrect = self._resource["reading_incorrect"]
        self.reading_max_streak = self._resource["reading_max_streak"]
        self.reading_current_streak = self._resource["reading_current_streak"]
        self.percentage_correct = self._resource["percentage_correct"]
        self.hidden = self._resource["hidden"]


class StudyMaterial(Resource, Subjectable):
    """
    Simple model holding information about Study Materials
    """

    resource = "study_material"

    def __init__(self, json_data, *args, **kwargs):
        super().__init__(json_data, *args, **kwargs)
        Subjectable.__init__(self, *args, **kwargs)
        self.created_at = parse8601(self._resource["created_at"])
        self.subject_id = self._resource["subject_id"]
        self.subject_type = self._resource["subject_type"]
        self.meaning_note = self._resource["meaning_note"]
        self.reading_note = self._resource["reading_note"]
        self.meaning_synonyms = self._resource["meaning_synonyms"]
        self.hidden = self._resource["hidden"]


class Lessons(object):
    def __init__(self, json_data, *args, **kwargs):
        self.subject_ids = json_data["subject_ids"]
        self.available_at = parse8601(json_data["available_at"])


class UpcomingReview(Subjectable):
    def __init__(self, json_data, *args, **kwargs):
        super().__init__(json_data, *args, **kwargs)
        Subjectable.__init__(self, *args, **kwargs)
        self.subject_ids = json_data["subject_ids"]
        self.available_at = parse8601(json_data["available_at"])


class Summary(Resource):
    resource = "report"

    def __init__(self, json_data, *args, **kwargs):
        super().__init__(json_data, *args, **kwargs)
        self.client = kwargs.get("client")
        # Note that there is only ever one lesson object, as per this forum thread https://community.wanikani.com/t/api-v2-alpha-documentation/18987
        self.lessons = Lessons(self._resource["lessons"][0])
        self.next_reviews_at = self._resource["next_reviews_at"]
        self.reviews = [
            UpcomingReview(review_json, client=self.client)
            for review_json in self._resource["reviews"]
        ]


class Review(Resource, Subjectable):
    resource = "review"

    def __init__(self, json_data, *args, **kwargs):
        super().__init__(json_data, *args, **kwargs)
        Subjectable.__init__(self, *args, **kwargs)
        self.created_at = parse8601(self._resource["created_at"])
        self.assignment_id = self._resource["assignment_id"]
        self.subject_id = self._resource["subject_id"]
        self.starting_srs_stage = self._resource["starting_srs_stage"]
        self.ending_srs_stage = self._resource["ending_srs_stage"]
        self.incorrect_meaning_answers = self._resource["incorrect_meaning_answers"]
        self.incorrect_reading_answers = self._resource["incorrect_reading_answers"]


class LevelProgression(Resource):
    resource = "level_progression"

    def __init__(self, json_data, *args, **kwargs):
        super().__init__(json_data, *args, **kwargs)
        self.created_at = parse8601(self._resource["created_at"])
        self.level = self._resource["level"]
        self.unlocked_at = parse8601(self._resource["unlocked_at"])
        self.started_at = parse8601(self._resource["started_at"])
        self.passed_at = parse8601(self._resource["passed_at"])
        self.completed_at = parse8601(self._resource["completed_at"])


def parse8601(time_field):
    if time_field:
        return dateutil.parser.parse(time_field)
    else:
        return None


resources = {
    UserInformation.resource: UserInformation,
    Assignment.resource: Assignment,
    Review.resource: Review,
    ReviewStatistic.resource: ReviewStatistic,
    LevelProgression.resource: LevelProgression,
    StudyMaterial.resource: StudyMaterial,
    Reset.resource: Reset,
    Kanji.resource: Kanji,
    Vocabulary.resource: Vocabulary,
    Radical.resource: Radical,
    Summary.resource: Summary,
    Page.resource: Page,
}


def factory(resource_json, *args, **kwargs):
    try:
        return resources[resource_json["object"]](resource_json, *args, **kwargs)
    except KeyError:
        raise UnknownResourceException(
            "We have no clue how to handle resource of type: {}".format(
                resource_json["object"]
            )
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `wanikani_api` package."""

from wanikani_api.client import Client
from wanikani_api.models import Subject, UserInformation
from tests.utils.utils import (
    mock_user_info,
    mock_subjects,
    mock_assignments,
    mock_review_statistics,
    mock_study_materials,
    mock_summary,
    mock_reviews,
    mock_level_progressions,
    mock_resets,
    mock_subjects_with_cache,
    mock_single_subject,
)


def test_client_can_get_user_information(requests_mock):
    mock_user_info(requests_mock)

    client = Client("v2_api_key")

    user = client.user_information()
    assert isinstance(user, UserInformation)


def test_client_can_get_subjects(requests_mock):
    mock_subjects(requests_mock)

    client = Client("v2_api_key")

    subjects = client.subjects()
    assert len(subjects.current_page.data) > 0
    assert subjects.current_page.data[0].resource in ["vocabulary", "kanji", "radical"]


def test_client_can_get_assignments(requests_mock):
    mock_assignments(requests_mock)

    client = Client("v2_api_key")

    assignments = client.assignments()

    assert len(assignments.current_page.data) > 0


def test_client_can_get_review_statistics(requests_mock):
    mock_review_statistics(requests_mock)
    client = Client("v2_api_key")

    review_statistics = client.review_statistics()
    assert len(review_statistics.current_page.data) > 0


def test_client_can_get_study_materials(requests_mock):
    mock_study_materials(requests_mock)
    client = Client("v2_api_key")

    study_materials = client.study_materials()
    assert len(study_materials.current_page.data) > 0


def test_client_can_get_summary(requests_mock):
    mock_summary(requests_mock)
    client = Client("v2_api_key")

    summary = client.summary()
    assert summary.lessons is not None
    assert summary.reviews is not None


def test_client_can_get_reviews(requests_mock):
    mock_reviews(requests_mock)
    client = Client("v2_api_key")

    reviews = client.reviews()
    assert len(reviews.current_page.data) > 0


def test_client_can_get_level_progression(requests_mock):
    mock_level_progressions(requests_mock)
    client = Client("v2_api_key")

    progressions = client.level_progressions()
    assert len(progressions.current_page.data) > 0


def test_client_can_get_resets(requests_mock):
    mock_resets(requests_mock)
    client = Client("v2_api_key")

    resets = client.resets()
    assert len(resets.current_page.data) == 1


def test_singular_subject_retrieval(requests_mock):
    mock_single_subject(requests_mock)
    v2_api_key = "arbitrary_api_key"
    client = Client(v2_api_key)

    subject = client.subject(1)
    assert isinstance(subject, Subject)


def test_client_uses_cache(requests_mock):
    mock_subjects(requests_mock)
    mock_assignments(requests_mock)
    v2_api_key = "arbitrary_api_key"
    client = Client(v2_api_key, subject_cache_enabled=True)
    assignments = client.assignments()
    for ass in assignments:
        print(ass.subject.level)  # in theory here, if we have _not_ cached

    assert requests_mock.call_count == 2
    history = requests_mock.request_history
    assert "subjects" in history[0].url
    assert "assignments" in history[1].url


def test_etag_cache_decorator_works(mocker, requests_mock):
    mock_subjects_with_cache(requests_mock)
    v2_api_key = "arbitrary_api_key"
    client = Client(v2_api_key)

    mocker.spy(client, "_fetch_result_from_cache")
    subjects = client.subjects()
    cached_subjects = client.subjects()
    assert client._fetch_result_from_cache.call_count == 1



import re

from wanikani_api import constants
from .response_mocks import *


def mock_subjects(requests_mock):
    requests_mock.get(
        re.compile(constants.SUBJECT_ENDPOINT),
        json=SUBJECTS_PAGE,
        headers={"Etag": "abc123"},
    )


def mock_single_subject(requests_mock):
    requests_mock.get(
        re.compile(constants.SINGLE_SUBJECT_ENPOINT),
        json=SINGLE_SUBJECT,
        headers={"Etag": "abc123"},
    )


def mock_empty_subjects(requests_mock):
    requests_mock.get(
        re.compile(constants.SUBJECT_ENDPOINT),
        json=EMPTY_SUBJECTS_PAGE,
        headers={"Etag": "abc123"},
    )


# When making multiple calls to the subject endpoint, only answer with real data once, then just return a 304.
def mock_subjects_with_cache(requests_mock):
    requests_mock.register_uri(
        "GET",
        re.compile(constants.SUBJECT_ENDPOINT),
        [
            {"json": SUBJECTS_PAGE, "status_code": 200, "headers": {"Etag": "abc123"}},
            {"json": None, "status_code": 304},
        ],
    )


def mock_user_info(requests_mock):
    requests_mock.get(
        re.compile(constants.USER_ENDPOINT),
        json=USER_INFORMATION,
        headers={"Etag": "abc123"},
    )


def mock_assignments(requests_mock):
    requests_mock.get(
        re.compile(constants.ASSIGNMENT_ENDPOINT),
        json=ASSIGNMENTS_PAGE,
        headers={"Etag": "abc123"},
    )


def mock_review_statistics(requests_mock):
    requests_mock.get(
        re.compile(constants.REVIEW_STATS_ENDPOINT),
        json=REVIEW_STATISTICS_PAGE,
        headers={"Etag": "abc123"},
    )


def mock_level_progressions(requests_mock):
    requests_mock.get(
        re.compile(constants.LEVEL_PROGRESSIONS_ENDPOINT),
        json=LEVEL_PROGRESSIONS_PAGE,
        headers={"Etag": "abc123"},
    )


def mock_summary(requests_mock):
    requests_mock.get(
        re.compile(constants.SUMMARY_ENDPOINT), json=SUMMARY, headers={"Etag": "abc123"}
    )


def mock_resets(requests_mock):
    requests_mock.get(
        re.compile(constants.RESETS_ENDPOINT),
        json=RESETS_PAGE,
        headers={"Etag": "abc123"},
    )


def mock_reviews(requests_mock):
    requests_mock.get(
        re.compile(constants.REVIEWS_ENDPOINT),
        json=REVIEWS_PAGE,
        headers={"Etag": "abc123"},
    )


def mock_study_materials(requests_mock):
    requests_mock.get(
        re.compile(constants.STUDY_MATERIALS_ENDPOINT),
        json=STUDY_MATERIALS_PAGE,
        headers={"Etag": "abc123"},
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `wanikani_api` package."""

from wanikani_api.client import Client
from wanikani_api.models import Subject, UserInformation
from tests.utils.utils import (
    mock_user_info,
    mock_subjects,
    mock_assignments,
    mock_review_statistics,
    mock_study_materials,
    mock_summary,
    mock_reviews,
    mock_level_progressions,
    mock_resets,
    mock_subjects_with_cache,
    mock_single_subject,
)

api_key = "2510f001-fe9e-414c-ba19-ccf79af40060"
def test_real_connection_to_wanikani_user_information():

    client = Client(api_key)

    user = client.user_information()
    assert isinstance(user, UserInformation)


def test_real_connection_to_subjects():
    client = Client(api_key)
    subjects = client.subjects()
    assert len(subjects.current_page.data) > 0
    assert subjects.current_page.data[0].resource in ["vocabulary", "kanji", "radical"]

def test_real_connection_to_assignments():
    client = Client(api_key)
    assignments = client.assignments()
    assert len(assignments.current_page.data) > 0


def test_real_connection_to_review_statistics():
    client = Client(api_key)
    review_statistics = client.review_statistics()
    assert len(review_statistics.current_page.data) > 0


def test_real_connection_to_study_materials():
    client = Client(api_key)
    study_materials = client.study_materials()
    assert len(study_materials.current_page.data) > 0

def test_real_connection_to_summary():
    client = Client(api_key)
    summary = client.summary()
    assert summary.lessons is not None
    assert summary.reviews is not None


def test_real_connection_to_reviews():
    client = Client(api_key)
    reviews = client.reviews()
    assert len(reviews.current_page.data) > 0


def test_real_connection_to_level_progression():
    client = Client(api_key)
    progressions = client.level_progressions()

def test_real_connection_to_resets():
    client = Client(api_key)
    resets = client.resets()
    assert len(resets.current_page.data) == 0

def test_real_connection_to_singular_subject():
    client = Client(api_key)
    subject = client.subject(1)
    assert isinstance(subject, Subject)

def test_client_uses_cache(requests_mock):
    mock_subjects(requests_mock)
    mock_assignments(requests_mock)
    v2_api_key = "arbitrary_api_key"
    client = Client(v2_api_key, subject_cache_enabled=True)
    assignments = client.assignments()
    for ass in assignments:
        print(ass.subject.level)  # in theory here, if we have _not_ cached

    assert requests_mock.call_count == 2
    history = requests_mock.request_history
    assert "subjects" in history[0].url
    assert "assignments" in history[1].url


def test_etag_cache_decorator_works(mocker, requests_mock):
    mock_subjects_with_cache(requests_mock)
    v2_api_key = "arbitrary_api_key"
    client = Client(v2_api_key)

    mocker.spy(client, "_fetch_result_from_cache")
    subjects = client.subjects()
    cached_subjects = client.subjects()
    assert client._fetch_result_from_cache.call_count == 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `wanikani_api` package."""
import datetime

import pytest
import requests

from tests.utils.utils import mock_subjects, mock_assignments
from wanikani_api.client import Client
from wanikani_api.constants import ROOT_WK_API_URL


@pytest.fixture
def url_builder():
    from wanikani_api.url_builder import UrlBuilder

    return UrlBuilder(ROOT_WK_API_URL)


def test_url_builder_with_no_parameters(url_builder):
    expected = "https://api.wanikani.com/v2/user"
    actual = url_builder.build_wk_url("user")
    assert actual == expected


def test_url_builder_with_single_string_parameter(url_builder):
    expected = "https://api.wanikani.com/v2/subjects?types=vocabulary"
    actual = url_builder.build_wk_url("subjects", {"types": "vocabulary"})
    assert actual == expected


def test_url_builder_with_single_array_parameter(url_builder):
    expected = "https://api.wanikani.com/v2/subjects?types=vocabulary,kanji,radicals"
    actual = url_builder.build_wk_url(
        "subjects", {"types": ["vocabulary", "kanji", "radicals"]}
    )
    assert actual == expected


def test_url_builder_with_multiple_string_parameter(url_builder):
    expected = "https://api.wanikani.com/v2/subjects?slugs=女&types=vocabulary"
    actual = url_builder.build_wk_url("subjects", {"types": "vocabulary", "slugs": "女"})
    assert actual == expected


def test_url_builder_with_single_integer_parameter(url_builder):
    expected = "https://api.wanikani.com/v2/subjects?ids=1"
    actual = url_builder.build_wk_url("subjects", {"ids": 1})
    assert actual == expected


def test_subject_parameters_are_properly_converted(requests_mock):
    mock_subjects(requests_mock)
    v2_api_key = "arbitrary_api_key"
    client = Client(v2_api_key)

    client.subjects(ids=[1, 2, 3], hidden=False, slugs=["abc", "123"])

    assert requests_mock.call_count == 1
    assert (
        requests_mock.request_history[0].url
        == "https://api.wanikani.com/v2/subjects?hidden=false&ids=1,2,3&slugs=abc,123"
    )


def test_parameters_convert_datetime_to_string_correctly(requests_mock):
    mock_subjects(requests_mock)
    v2_api_key = "arbitrary_api_key"
    client = Client(v2_api_key)
    now = datetime.datetime.now()

    client.subjects(updated_after=now)

    assert requests_mock.call_count == 1
    assert (
        requests_mock.request_history[0].url
        == "https://api.wanikani.com/v2/subjects?updated_after=" + now.isoformat()
    )


def test_assignment_parameters_are_properly_converted(requests_mock):
    mock_assignments(requests_mock)
    v2_api_key = "arbitrary_api_key"
    client = Client(v2_api_key)

    client.assignments(ids=[1, 2, 3], hidden=False, srs_stages=[0, 1, 2])

    assert requests_mock.call_count == 1
    assert (
        requests_mock.request_history[0].url
        == "https://api.wanikani.com/v2/assignments?hidden=false&ids=1,2,3&srs_stages=0,1,2"
    )
